
NUEVO = "nuevo"

class Bill:

    def __init__(self, document_type, document, raw_code, amount, errors):
        self.errors = False
        self.skip = False

        if document_type == "Factura":
            document_type = "Factura de Venta"

        document_type_and_number = "%s N° %s" % (document_type, document)

        self.document_type = document_type
        self.document = document
        self.raw_code = raw_code
        self.amount = amount
        self.document_type_and_number = document_type_and_number

        if not raw_code:
            self.append_error(errors, "Factura sin descripción. Documento: %s" % document)
            return

        if len(raw_code.split("-")) < 4:
            self.append_error(errors, "Factura con código incorrecto (%s). Documento: %s" % (raw_code, document))
            return

        _, _, self.sales_order, self.payment_code, *_ = raw_code.split("-")

        if "de" in self.payment_code:
            self.skip = True
            return
        '''
        if self.sales_order == "NO":
            self.skip = True
            return
        '''

        if self.sales_order == "NO":
            self.skip = True
            return


        if not self.sales_order:
            error = "Factura sin orden de compra. Documento: %s" % document
            self.append_error(errors, error)
            return

        self.version = NUEVO

        if self.sales_order != "NO" and len(self.sales_order) != 5:
            error = "Factura con orden de compra errónea (%s). Documento: %s" % (self.sales_order, self.document)            
            print(error)
            self.append_error(errors, error)
            return

    def append_error(self, errors, error):
        self.errors = True
        errors.append(error)

    def is_duplicated_bill(self, bills, errors):
        if self.document_type_and_number in bills and self.version == NUEVO:
            errors.append("Orden de compra repetida. Documento: %s. Orden de compra: %s" % (self.document, self.sales_order))
            return True
        return False

    def should_skip(self):
        return self.errors or self.skip

    def get_document_type_and_number(self):
        return self.document_type_and_number

    def get_amount(self):
        return self.amount



from credisur.models import Bill


def bill_row_extractor(results, unpacker):

    errors = results.get_errors()
    bills = results.get_bills()

    document_type = unpacker.get_value_at(3)
    document = unpacker.get_value_at(4)
    raw_code = unpacker.get_value_at(11)
    amount = unpacker.get_value_at(18)

    bill = Bill(document_type, document, raw_code, amount, errors)

    if bill.should_skip():
        return

    if bill.is_duplicated_bill(bills, errors):
        return

    results.append_bill(bill)



from . import ExcelWriter
from credisur.excelbuilder import BasicBuilder

class ExcelBasicComposer:

    def __init__(self, filename, collection, columns_configuration, sheetname):
        self.writer = ExcelWriter(filename)
        builder = BasicBuilder(collection, columns_configuration)
        self.writer.build_sheet(sheetname, builder.build_sheet_data())

    def save(self):
        self.writer.save()


from .rowunpacker import *


class TableUnpacker:
    def __init__(self, sheet):
        self.sheet = sheet

    def get_row_unpacker(self, row):
        return RowUnpacker(self.sheet, row)

    def read_rows(self, initial_row):
        for row_num in range(initial_row, self.sheet.max_row + 1):
            yield self.get_row_unpacker(row_num)



import openpyxl


class ExcelReader:
    def __init__(self, filename):
        self.workbook = openpyxl.load_workbook(filename=filename, data_only=True)

    def get_sheet(self, sheetname):
        return self.workbook.get_sheet_by_name(sheetname)





class DebtorsReport:
    
    def generate_report(self, accounts_to_collect, customers, first_day_of_last_month):

        result = self.build_result()

        for collection_account, accounts in accounts_to_collect.items():
            
            for acct in accounts:
                due_payments = acct['missing_payments']
                last_collection_date = acct['last_collection_as_date']            
                
                if not acct['version'] == "nuevo":
                    continue

                if self.is_debtor(due_payments, last_collection_date, first_day_of_last_month):

                    customer = self.get_account_customer(acct, customers)

                    debtor = self.build_debtor(acct, customer, collection_account)
                
                    result.append(debtor)

        return result


    def get_account_customer(self, acct, customers):
        return customers[acct['customer']]


    def is_debtor(self, due_payments, last_collection_date, first_day_of_last_month):
        if due_payments >= 3:
            return True
        
        if last_collection_date:
            return last_collection_date < first_day_of_last_month

        return False

    def build_result(self):
        return list()    


    def build_debtor(self, acct, customer, collection_account):
        # Ciudad	Cliente	Dirección	Teléfono	Fecha de Compra	Fecha de Vencimiento	
        # Orden de Compra	Última Cobranza	Cuotas	Cuotas a pagar	Valor de cuota	
        # Monto total a cobrar	Descripción	CBU

        debtor = {}
        debtor['city'] = customer['city']
        debtor['customer'] = acct['customer']
        debtor['address'] = customer['address']
        debtor['phone'] = customer['phone']
        debtor['date_of_purchase'] = acct['date_of_purchase']
        debtor['due_date'] = acct['due_date']
        debtor['order'] = acct['order']
        debtor['last_collection'] = acct['last_collection']        
        debtor['plan'] = acct['plan']
        debtor['missing_payments'] = acct['missing_payments']
        debtor['payment'] = acct['payment']
        debtor['amount_to_collect'] = acct['amount_to_collect']
        debtor['full_description'] = acct['full_description']
        debtor['cbu'] = customer['cbu']
        debtor['collection_account'] = collection_account

        return debtor


import openpyxl

def read_debits_excel(excel_path):
    workbook = openpyxl.load_workbook(
        filename=excel_path, read_only=True, data_only=True)
    # workbook.get_sheet_by_name(sheetname)

    mapper_func = sheet_builder(workbook)
    bank_sheets = map(mapper_func, list(filter(is_bank_sheet, workbook.sheetnames)))

    return bank_sheets


def is_bank_sheet(sheet_name):
    return sheet_name.startswith("Banco")

def sheet_builder(workbook):
    def mapper(sheetname):
        return workbook.get_sheet_by_name(sheetname)

    return mapper


def get_no_payment_due_columns():
    config_list = list()
    config_list.append(("A", "Ciudad", 'city'))
    config_list.append(("B", "Cliente", 'customer'))
    config_list.append(("C", "Orden de Compra", 'sales_order'))
    config_list.append(("D", "Dirección", 'address'))
    config_list.append(("E", "Monto Vencido", 'past_due_debt'))
    config_list.append(("F", "Monto Cobrado", 'paid_amount'))
    config_list.append(("G", "Primer Vencimiento", 'due_date'))
    config_list.append(("H", "Código", 'raw_code'))
    return config_list

'''
"%s - %s (orden: %s) - %s. Monto vencido: %s. Monto cobrado: %s. Primer vencimiento: %s" % (

    "city": self.city,
"customer": self.customer,
"sales_order": self.sales_order,
"adddress": self.address or 'Sin dirección',
"past_due_debt": self.past_due_debt,
"paid_amount": self.paid_amount,
"due_date": self.due_date.strftime("%d/%m/%Y")
'''


class HashOfLists:
    def __init__(self):
        self.hash = {}

    def append(self, key, item):
        if not key in self.hash:
            self.hash[key] = []

        self.hash[key].append(item)

    def items(self):
        for key, list in self.hash.items():
            yield key, list

    def __getitem__(self, key):
        return self.hash[key]

    def keys(self):
        return self.hash.keys()

    def __contains__(self, key):
        return key in self.hash



def get_bank_debit_columns():
    config_list = list()
    config_list.append(("A", "Tipo de Novedad", 'event_type'))
    config_list.append(("B", "CBU Bloque 1", 'cbu1'))
    config_list.append(("C", "CBU Bloque 2", 'cbu2'))
    config_list.append(("D", "Identificación del cliente", 'customer'))
#    config_list.append(("E", "Ref. del débito", 'ref_credit'))
    config_list.append(("E", "Importe", 'amount'))
    config_list.append(("G", "Deuda total", 'total_payment'))
    config_list.append(("H", "Última Cobranza", 'last_collection'))
    return config_list


class RowUnpacker:
    def __init__(self, sheet, row_num):
        self.sheet = sheet
        self.row_num = row_num

    def get_value_at(self, col):
        return self.sheet.cell(row=self.row_num, column=col).value



def customer_row_extractor(results, unpacker):
    result = dict()

    name = unpacker.get_value_at(1)

    result['address'] = unpacker.get_value_at(8)
    result['city'] = unpacker.get_value_at(33) # Cambio de Xubio 2019 09 (de 33 a 32) - 2do cambio en 2019 09 (de 32 a 33)
    result['phone'] = unpacker.get_value_at(12)
    result['cbu'] = unpacker.get_value_at(6) # Cambio de Xubio 2019 09 (de 5 a 6)

    results[name] = result


from functools import reduce


def calculate_last_line(header_data, debits_data):
    """

    :type header_data: credisur.debitsgenerator.parsers.parsedebitsheaders.Header
    """
    monetary_records = len(debits_data)
    total_amount = reduce(lambda x, y: x + y, map(lambda x: x.monto_numero, debits_data))
    process_date = header_data.fecha_proceso

    return LastLine(process_date, monetary_records, total_amount)

def format_amount(amount):
    return str(round(amount * 100)).zfill(10)


def format_total_number_of_records(number):
    return str(number).zfill(10)


def format_number_of_records(number):
    return str(number).zfill(7)


class LastLine:

    def __init__(self, fecha_proceso, registros_monetarios, suma_importes_registros):
        # formato de registros: 7 caracteres y relleno con ceros
        # formato de registros TOTALES: 10 caracteres y relleno con ceros
        # formato de montos: 10 caracteres, en centavos y relleno con ceros
        # formato fecha: ddmmaaaa
        # FILLER 1: del 34 al 103 (70 chars en blanco)
        # FILLER 2: del 114 al 250 (137 chars en blanco)

        self._tipo_de_novedad = "T"
        self._registros_monetarios = format_number_of_records(registros_monetarios)
        self._registros_no_monetarios = format_number_of_records(0)
        self._registros_totales = format_total_number_of_records(registros_monetarios)
        self._fecha_proceso = fecha_proceso
        self._suma_importes_registros = format_amount(suma_importes_registros)


    @property
    def tipo_de_novedad(self):
        return self._tipo_de_novedad

    @property
    def registros_totales(self):
        return self._registros_totales

    @property
    def registros_monetarios(self):
        return self._registros_monetarios

    @property
    def registros_no_monetarios(self):
        return self._registros_no_monetarios

    @property
    def fecha_proceso(self):
        return self._fecha_proceso

    @property
    def suma_importes_registros(self):
        return self._suma_importes_registros





class AccountReceivableExtractorResults:

    def __init__(self,
                 customers, bills,
                 calendar_ops, first_day_of_current_month, last_date_of_month,
                 collections, collections_for_customers,
                 list_of_advance_payments, customers_without_payments_due, customers_in_last_payment):

        self.accounts_to_collect =  {
            "C": [],
            "D": [],
            "I": []
        }

        self.customers = customers
        self.bills = bills

        self.calendar_ops = calendar_ops
        self.last_date_of_month = last_date_of_month
        self.first_day_of_current_month = first_day_of_current_month

        self.errors = []

        self.collections = collections
        self.collections_for_customers = collections_for_customers
        self.list_of_advance_payments = list_of_advance_payments


        self.customers_without_payments_due = customers_without_payments_due
        self.customers_in_last_payment = customers_in_last_payment


    def get_errors(self):
        return self.errors

    def get_calendar_ops(self):
        return self.calendar_ops

    def get_first_day_of_current_month(self):
        return self.first_day_of_current_month

    def get_last_date_of_month(self):
        return self.last_date_of_month

    def get_accounts_to_collect(self):
        return self.accounts_to_collect

    def get_customers(self):
        return self.customers

    def get_bills(self):
        return self.bills

    def get_collections(self):
        return self.collections

    def get_collections_for_customers(self):
        return self.collections_for_customers

    def get_list_of_advance_payments(self):
        return self.list_of_advance_payments

    def get_customers_without_payments_due(self):
        return self.customers_without_payments_due

    def get_customers_in_last_payment(self):
        return self.customers_in_last_payment

    def append_account_to_collect(self, account_receivable):
        account_to_collect = account_receivable.to_dict()

        if not account_to_collect['city'] or not account_to_collect['customer']:
            print("FALTA CIUDAD O CLIENTE:", account_to_collect['city'], account_to_collect['customer'])

        self.accounts_to_collect[account_to_collect['account']].append(account_to_collect)


import datetime

def parse_debits_headers(page_data):
    headers_mapping = map_headers(page_data)
    return Header(page_data.title, headers_mapping)

def map_headers(page_data):
    return {
        'cuit': page_data['B1'].value,
        'prestacion': page_data['B2'].value,
        'ref_debito': page_data['B3'].value,
        'primer_vencimiento': page_data['B4'].value,
        'fecha_proceso': page_data['B5'].value,
    }

class Header:

    def __init__(self, sheet_name, headers_map):
        self._sheet_name = sheet_name
        self._cuit = headers_map['cuit']
        self._prestacion = headers_map['prestacion']
        self._ref_debito = headers_map['ref_debito']
        self._primer_vencimiento = headers_map['primer_vencimiento']
        self._fecha_proceso = headers_map['fecha_proceso']

    def format_date(self, a_date):
        if type(a_date) is str:
            a_date = datetime.datetime.strptime(a_date, "%Y-%m-%d %h:%m:%s")
        return a_date.strftime("%d%m%Y")

    @property
    def sheet_name(self):
        return self._sheet_name

    @property
    def cuit(self):
        return self._cuit

    @property
    def prestacion(self):
        return self._prestacion

    @property
    def ref_debito(self):
        return self._ref_debito

    @property
    def primer_vencimiento(self):
        if not self._primer_vencimiento:
            return None

        return self.format_date(self._primer_vencimiento)

    @property
    def fecha_proceso(self):
        if not self._fecha_proceso:
            return None

        return self.format_date(self._fecha_proceso)


from .errorsiterator import ErrorsIterator

def validate_headers(header_data):
    errors = ErrorsIterator()
    sheet_name = header_data.sheet_name

    if header_data.cuit is None:
        errors.add_error("Falta el CUIT en solapa %s" % (sheet_name))

    if header_data.prestacion is None:
        errors.add_error("Falta la 'Prestación' en solapa %s" % (sheet_name))

    if header_data.prestacion is None:
        errors.add_error("Falta la 'Ref del débito' en solapa %s" % (sheet_name))

    if header_data.primer_vencimiento is None:
        errors.add_error("Falta la 'Fecha 1er vencimiento' en solapa %s" % (sheet_name))

    if header_data.fecha_proceso is None:
        errors.add_error("Falta la 'Fecha de proceso' en solapa %s" % (sheet_name))

    return errors



from setuptools import setup, find_packages

version = {}
exec(open('credisur/version.py').read(), version)


setup(name='credisuretl',
      version=version['__version__'],
      description='ETL for Felsim',
      url='http://github.com/fedegos/credisuretl',
      author='Federico Gosman',
      author_email='federico@equipogsm.com',
      license='MIT',
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'credisur = credisur.__main__:main'
          ]
      },
      install_requires=[
            'openpyxl==2.4.7',
            'xlrd',
            'time', 
            'python-dateutil==2.8.0'
      ],
      zip_safe=False)


# TODO: Agregar última cuota

def get_last_payment_columns():
    config_list = list()
    config_list.append(("A", "Ciudad", 'city'))
    config_list.append(("B", "Cliente", 'customer'))
    config_list.append(("C", "Dirección", 'address'))
    config_list.append(("D", "Última Cobranza", 'lastcollection'))
    config_list.append(("E", "Motivo", 'reason'))
    config_list.append(("F", "Valor última cuota", 'payment'))
    return config_list